/* Client for the BALANCER "load balancer" utility for GoServe servers.
   To use this, you must set several variables, including:
 *) The IP address of the main server, and the port to use.
    Load information is sent to this server:port combo.
 *) Frequency of transmission of this information
 *) Load info type -- average response time, average "busy" time, and
                      system busy time, dumb default

For details, see BALANCER.DOC.

*/

/* ********** BEGIN USER CHANGEABLE PARAMETERS **********/
/*Note: the following parameter can be overriden by command
        line options of the form:  
             d:\path>CPUloadC  avar=avalue avar=avalu
        where avar is Main_server, etc.
*/

/* the IP address of the "main" server. You can use either numeric or  name */
MAIN_SERVER='loopy'

/* Port to use.  MUST match the value of the TYPE_2_PORT variable 
   in the BALANCER.80 GoServe filter on the "main server". */
PORT=8802

/* Name of this server.  Must exactly match the name used by the
  main_server. If you leave this blank, this server's ip address
  will be used. 
    The crucial point is that it MUST match the name expected by the 
    main_server.  That is, if this subsidiary server is at FOO.BAR.NET, and a 
    main_server (on your sub-domain) is expecting FOO, then you MUST 
    set myname='FOO', not 'FOO.BAR.NET' (that is, no attempt is made 
    to match/resolve IP addresses; nor match name to numbers).  */
MYNAME='LOOPY'

/* Seconds to wait between "updates".  That is, updates on this server's
   load are sent to the main_server every "WAIT_SECONDS" seconds */
WAIT_SECONDS=60


/* Mode. Three values are understood: 1,2, or 3
    MODE=1 : Use the extract('responsetime') GoServe statistic
    MODE=2 : Use total bytes sent (in last wait_seconds seconds)
    MODE=3 : Use QPROCSTATUS (requires RXU)  
    MODE=4 : Real dumb default -- time it takes to count to 4000
  To use MODE1 or MODE2 you MUST launch this as a semi-permanent
  thread under GoServe (say, by using rexxlib's REXXTHREAD procedure). 
  Note: SRE-http users can use the CUSTOM_INITS parameter to "launch"
       this program.
  To use MODE3, RXU must be in your LIBPATH.
  If the selected MODE can not be used, them MODE=4 is used as a default.
*/
MODE=2 

/* verbose=0 - no status info, 1=some, 2=lots */
VERBOSE=2

/* usesay=1 == disable use of pmprintf for status reports */
USESAY=0

/* ********** END USER CHANGEABLE PARAMETERS **********/


parse arg copts

if strip(copts)='?' then do
 say " CPUloadC is the client for the BALANCER GoServe-load-balancer."
 say " It requires several parameters that may be set in the code, or may"
 say " be set with command line options. "
 say " To use command line options, start CPUloadC as: "
 say "  d:\path>CPUloadC  option1=avalue [... optionN=avalue ] "
 say "(note that each option=avalue pair should be seperated by a space) "
 say
 say "Valid options are: "
 say '   MAIN_SERVER : IP address of the "main server" '
 say '          PORT : Port that main_server is listening to '
 say '          MODE : How to compute load -- 1,2,3 or 4 '
 say '       VERBOSE : Level of status reporting (0=little, 1=some, or 2=lot) '
 say '        MYNAME : Name of this server (as expected by main_server) '
 say '  WAIT_SECONDS : Frequency of update (in seconds} '
 say '        USESAY : If=1, then use SAY (not PMPRINTF) for status reports '
 say
 say " For details, see the code, or see BALANCER.DOC "
 exit
end /* do */
say ' -------------'

do until copts="" 
    parse var copts a1 copts
    parse upper var a1 avar '=' aval ;aval=strip(aval); avar=strip(avar)
    select 
       when avar='MAIN_SERVER' then main_server=aval
       when avar='MODE' then mode=aval
       when avar='USESAY' then usesay=1
       when avar='VERBOSE'  then  verbose=aval
       when avar='MYNAME' then myname=aval
       when avar='PORT' then port=aval
       when  avar='WAIT_SECONDS' then wait_seconds=aval
       otherwise nop
    end  /* select */
end

/* rexxlib */
foo=rxfuncquery('rexxlibregister')
if foo=1 then do
 call rxfuncadd 'rexxlibregister','rexxlib', 'rexxlibregister'
 call rexxlibregister
end

no_pmprintf=rxfuncquery('pmprintf')
if usesay=1 then no_pmprintf=1  /* disable pmprintf */

call dosay "CPULoadC. Send sys-load info to  " main_server':'port
call dosay " (enter CPUloadC ? for instructions) "

call load_dll           /*load some dlls */

if verify(main_server,'0123456789.') <>0 then do
   rc=sockgethostbyname(main_server, "serv.0")  /* get dotaddress of server */
   if rc=0 then do;
      call dosay 'Unable to resolve 'main_server 
      exit
   end
   main_server=serv.0addr
   call dosay " Using numeric address: "main_server
end


if myname='' | myname=0 then myname=get_hostname()
call dosay " Using myname=" myname
/* see if desired mode is allowd */
call dosay " Using mode="mode
mode=use_mode(mode)

signal on error name err1 ; signal on syntax name err1

do forever              /* infinite loop */
  select  
     when mode=1 then  ares=extract('responsetime')
     when mode=2 then ares=extract('clients')+0.5
     when mode=3 then ares=cpu_load()
     otherwise do
        aa=time('r')
        igoo=0
        do geeb=1 to 4000
           arf=time('e')
           igoo=igoo+1
        end /* do */
        ares=time('e')
      end
  end  /* select */

  if verbose>1 then call dosay " Load value= "ares
  amess=myname' 'ares
  aa=send_results(main_server,port,amess)

  call syssleep wait_seconds
end /* do */


err1:
call dosay " Error at line " sigl "( "rc
exit



/**********************/
use_mode:Procedure expose no_pmprintf
parse arg amode


if amode=4 then return 4   /* always works */

signal on syntax name nogos ; signal on error name nogos ;
if amode=1 | amode=2 then do    /* see if 'extract' works */
  foo=extract('responsetime')
  foo2=extract('clients')
  return amode          /* if here, no error */
end

if amode=3 then do  /* see if qprocstatus is available */
/* requires rxu */
  if rxfuncquery('rxuinit')=1 then do
     call rxfuncadd 'rxuinit','rxu','rxuinit'
     call rxuinit
  end
  if rxfuncquery('rxuinit')=1 then do
      call dosay " Unable to find RXU -- using mode 4 "
      return 4
   end
   return amode
end /* do */

nogos:
call dosay " Could not find GoServe procedures -- using mode 4"
return 4


/***********************************/
send_results:procedure expose verbose no_pmprintf
parse arg dotserver,port,message
crlf='0d0a'x                        
family  ='AF_INET'

gosaddr.0family=family                  /* set up address */
gosaddr.0port  =port
gosaddr.0addr  =dotserver

gosock = SockSocket(family, "SOCK_STREAM", "IPPROTO_TCP")
rc = SockConnect(gosock,"gosaddr.0")

if rc<0 then do;
    call dosay ' Unable to connect to 'dotserver
    rc = SockClose(gosock)
    return 0
end

rc = SockSend(gosock, message)
if rc<=0 then 
  call dosay " Problem with SockSend "
if rc>0 & verbose>1 then 
  call dosay " Send to "dotserver" : " message


rc = SockClose(gosock)

return 1


/* --- Load the function library, if necessary --- */
load_dll:

/* socket procedures */
if RxFuncQuery("SockLoadFuncs")=1 then do      /* already there */
   call RxFuncAdd "SockLoadFuncs","rxSock","SockLoadFuncs"
   call SockLoadFuncs
end
if rxfuncquery('sockloadfuncs')=1 then do
   call dosay " Unable to find RexxSock "
   exit
end /* do */

/* rexxlib */
foo=rxfuncquery('rexxlibregister')
if foo=1 then do
 call rxfuncadd 'rexxlibregister','rexxlib', 'rexxlibregister'
 call rexxlibregister
end

/* Load up advanced REXX functions */
foo=rxfuncquery('sysloadfuncs')
if foo=1 then do
  call RxFuncAdd 'SysLoadFuncs', 'RexxUtil', 'SysLoadFuncs'
  call SysLoadFuncs
end
if rxfuncquery('sysloadfuncs')=1 then do
   call dosay " Unable to find RexxUtil "
   exit
end /* do */


return 1


/*)((((((((((((((((((((((((((((((( */
/* cpu load ,in time slices used per second */
cpu_load:procedure expose no_pmprintf

oldtotu=0; oldtots=0


tot1=get_proctime(1,a1)  /* 1=u ,2=s, 3=both */
foo=time('r')
call syssleep 1
foo=time('e')
tot2=get_proctime(2,a1)

/* compute difference */
asum=0 ;ns=0
do mm=1 to max(tot1,tot2)
/* skip thread that's new or just died */
   if symbol('TIMERS.1.'MM)<>'VAR' | symbol('TIMERS.2.'MM)<>'VAR' then iterate
   ns=ns+1
   diff=timers.2.mm-timers.1.mm
   if diff>0 then asum=asum+diff  /* drop die/alive threads within the second */
end /* do */
flops=asum/(16*foo)
return flops            /* used time slices per second */


/***/
/* get total process time.
  Return total time, and set timers.ith.n; n=1... # threads */
get_Proctime:procedure expose timers. no_pmprintf
parse arg ith,gets

/* use qproc status to find system utilization */
foo=rxqprocstatus('slist.')
nu=0 ; sys=0
do mm=1 to slist.0p.0
  nts=slist.0p.mm.0t.0
  do nn=1 to nts
    oof=x2d(slist.0p.mm.0t.nn.2)
    select
      when gets=1 then timers.ith.oof=slist.0p.mm.0t.nn.5
      when gets=2 then timers.ith.oof=slist.0p.mm.0t.nn.6
      otherwise timers.ith.oof=slist.0p.mm.0t.nn.6+slist.0p.mm.0t.nn.5
    end
  end
end /* do */

return slist.0g.1  /* # of threads */


/* get the hostname (aa.bb.cc) for this machine */
get_hostname: procedure
    do queued(); pull .; end                   /* flush */
    address cmd '@hostname'  '| rxqueue'    
    parse pull hostname                        
    return hostname

/*****/
/* say or pmprintf */
dosay:
parse arg amm
if no_pmprintf=1 then
  say 'CPUloadC. 'amm
else
  call pmprintf('CPUloadC. 'amm)
return 0
